import logging
import uuid
from contextlib import asynccontextmanager, contextmanager
from contextvars import ContextVar, Token
from logging import Logger
from typing import (
    Optional,
    Dict,
    Any,
    override,
    Iterator,
    AsyncIterator,
    List,
    Union,
    Mapping,
    ClassVar,
)

from opentelemetry.metrics import NoOpCounter, NoOpUpDownCounter, NoOpHistogram

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.metrics.telemetry_counter import TelemetryCounter
from helixtelemetry.telemetry.metrics.telemetry_histogram_counter import (
    TelemetryHistogram,
)
from helixtelemetry.telemetry.metrics.telemetry_up_down_counter import (
    TelemetryUpDownCounter,
)
from helixtelemetry.telemetry.providers.telemetry import Telemetry
from helixtelemetry.telemetry.spans.console_telemetry_span_wrapper import (
    ConsoleTelemetrySpanWrapper,
)
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.console_telemetry_history_item import (
    ConsoleTelemetryHistoryItem,
)
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class ConsoleTelemetry(Telemetry):
    telemetry_provider: ClassVar[str] = "ConsoleTelemetry"

    _CONTEXT_KEY = "current_context"
    _telemetry_history: List[ConsoleTelemetryHistoryItem] = []
    _current_context_variable: ContextVar[Optional[TelemetryParent]] = ContextVar(
        _CONTEXT_KEY, default=None
    )

    # _current_thread_context_variable: TypedThreadLocal[Optional[TelemetryContext]] = (
    #     TypedThreadLocal()
    # )

    def __init__(
        self,
        *,
        telemetry_context: TelemetryContext,
        log_level: Optional[Union[int, str]],
    ) -> None:
        super().__init__(
            telemetry_context=telemetry_context,
            log_level=log_level,
        )
        self._logger: Logger = logging.getLogger(__name__)
        # get_logger sets the log level to the environment variable LOGLEVEL if it exists
        self._logger.setLevel(log_level or "INFO")

    # noinspection PyUnusedLocal
    @classmethod
    def add_telemetry_history_item(
        cls, *, parent: TelemetryParent | None, item: ConsoleTelemetryHistoryItem
    ) -> None:
        # otherwise add it to the root
        cls._telemetry_history.append(item)

    @contextmanager
    @override
    def trace(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> Iterator[TelemetrySpanWrapper]:

        # read the current value of the context variable
        current_value: TelemetryParent | None = self._current_context_variable.get()

        # if parent is passed then use the trace_id and span_id from the parent
        if telemetry_parent is None:
            telemetry_parent = current_value

        # if parent is still None then create a new trace_id and span_id
        if telemetry_parent is None:
            telemetry_parent = TelemetryParent(
                trace_id=str(uuid.uuid4()),
                span_id=str(uuid.uuid4()),
                name=name,
                attributes=attributes,
                telemetry_context=self._telemetry_context,
            )

        token: Token[TelemetryParent | None] | None = (
            self._current_context_variable.set(telemetry_parent)
        )
        current_value = self._current_context_variable.get()
        assert telemetry_parent is not None
        try:
            self.add_telemetry_history_item(
                parent=current_value,
                item=ConsoleTelemetryHistoryItem.from_telemetry_context(
                    name=name,
                    telemetry_context=self._telemetry_context,
                    telemetry_parent=telemetry_parent,
                ),
            )
            yield ConsoleTelemetrySpanWrapper(
                name=name,
                attributes=attributes,
                telemetry_context=self._telemetry_context,
                telemetry_parent=telemetry_parent,
            )
        finally:
            if token is not None:
                self._current_context_variable.reset(token)

    @asynccontextmanager
    @override
    async def trace_async(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> AsyncIterator[TelemetrySpanWrapper]:
        # read the current value of the context variable
        current_value: TelemetryParent | None = self._current_context_variable.get()

        # if parent is passed then use the trace_id and span_id from the parent
        if telemetry_parent is None:
            telemetry_parent = current_value

        # if parent is still None then create a new trace_id and span_id
        if telemetry_parent is None:
            telemetry_parent = TelemetryParent(
                trace_id=str(uuid.uuid4()),
                span_id=str(uuid.uuid4()),
                name=name,
                attributes=attributes,
                telemetry_context=self._telemetry_context,
            )

        token: Token[TelemetryParent | None] | None = (
            self._current_context_variable.set(telemetry_parent)
        )
        current_value = self._current_context_variable.get()
        assert telemetry_parent is not None
        try:
            self.add_telemetry_history_item(
                parent=current_value,
                item=ConsoleTelemetryHistoryItem.from_telemetry_context(
                    name=name,
                    telemetry_context=self._telemetry_context,
                    telemetry_parent=telemetry_parent,
                ),
            )
            yield ConsoleTelemetrySpanWrapper(
                name=name,
                attributes=attributes,
                telemetry_context=self._telemetry_context,
                telemetry_parent=telemetry_parent,
            )
        finally:
            if token is not None:
                self._current_context_variable.reset(token)

    @override
    def track_exception(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        pass

    @override
    async def track_exception_async(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        pass

    @override
    async def flush_async(self) -> None:
        pass

    @override
    async def shutdown_async(self) -> None:
        pass

    def __getstate__(self) -> Dict[str, Any]:
        # Exclude certain properties from being pickled otherwise they cause errors in pickling
        return {k: v for k, v in self.__dict__.items()}

    @override
    def get_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        add_metadata: Optional[List[str]] = None,
    ) -> TelemetryCounter:
        """
        Get a counter metric
        :param name: Name of the counter
        :param unit: Unit of the counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryCounter(
            counter=NoOpCounter(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def get_up_down_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryUpDownCounter:
        """
        Get an up_down_counter metric

        :param name: Name of the up_down_counter
        :param unit: Unit of the up_down_counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryUpDownCounter(
            counter=NoOpUpDownCounter(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def get_histogram(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryHistogram:
        """
        Get a histograms metric

        :param name: Name of the histograms
        :param unit: Unit of the histograms
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryHistogram(
            histogram=NoOpHistogram(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )



from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.providers.console_telemetry import ConsoleTelemetry
from helixtelemetry.telemetry.providers.null_telemetry import NullTelemetry
from helixtelemetry.telemetry.providers.open_telemetry import OpenTelemetry


def register() -> None:
    """
    Register the telemetry classes with the telemetry factory
    """

    TelemetryFactory.register_telemetry_class(
        name=NullTelemetry.telemetry_provider, telemetry_class=NullTelemetry
    )
    TelemetryFactory.register_telemetry_class(
        name=ConsoleTelemetry.telemetry_provider, telemetry_class=ConsoleTelemetry
    )

    TelemetryFactory.register_telemetry_class(
        name=OpenTelemetry.telemetry_provider,
        telemetry_class=OpenTelemetry,
    )



import time

import pytest

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.providers.open_telemetry import OpenTelemetry


@pytest.mark.skip(reason="This test is for manual testing only")
async def test_open_telemetry_async() -> None:
    # Initialize telemetry
    telemetry_context = TelemetryContext(
        provider=OpenTelemetry.telemetry_provider,
        service_name="example-service",
        environment="development",
        attributes=None,
        log_level=None,
        instance_name="main-operation",
        service_namespace="MyNamespace",
    )
    telemetry = OpenTelemetry(telemetry_context=telemetry_context, log_level="DEBUG")

    # Using as a context manager
    with telemetry.trace(
        name="foo_operation",
        attributes={"key": "value", "key2": "foo"},
        telemetry_parent=None,
    ):
        print("Performing operation")
        time.sleep(0.05)
        with telemetry.trace(
            name="bar_operation", attributes={"key2": "bar"}, telemetry_parent=None
        ):
            print("Performing sub-operation")
            time.sleep(0.05)
            telemetry.add_event("sub-operation_completed", {"key": "value"})
            telemetry.track_exception(ValueError("Test2 exception"))
        telemetry.add_event("operation_completed", {"key": "value"})

        # Shutdown telemetry
        await telemetry.shutdown_async()



import logging
from contextlib import contextmanager, asynccontextmanager
from logging import Logger
from typing import (
    Optional,
    Dict,
    Any,
    Iterator,
    AsyncIterator,
    override,
    Mapping,
    Union,
    ClassVar,
    List,
)

from opentelemetry.metrics import NoOpCounter, NoOpUpDownCounter, NoOpHistogram

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.metrics.telemetry_counter import TelemetryCounter
from helixtelemetry.telemetry.metrics.telemetry_histogram_counter import (
    TelemetryHistogram,
)
from helixtelemetry.telemetry.metrics.telemetry_up_down_counter import (
    TelemetryUpDownCounter,
)
from helixtelemetry.telemetry.providers.telemetry import Telemetry
from helixtelemetry.telemetry.spans.null_telemetry_span_wrapper import (
    NullTelemetrySpanWrapper,
)
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class NullTelemetry(Telemetry):
    telemetry_provider: ClassVar[str] = "NullTelemetry"

    def __init__(
        self,
        *,
        telemetry_context: TelemetryContext,
        log_level: Optional[Union[int, str]],
    ) -> None:
        super().__init__(
            telemetry_context=telemetry_context,
            log_level=log_level,
        )
        self._logger: Logger = logging.getLogger(__name__)
        # get_logger sets the log level to the environment variable LOGLEVEL if it exists
        self._logger.setLevel(log_level or "INFO")

    @override
    def track_exception(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        pass

    @override
    async def track_exception_async(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        pass

    @override
    async def flush_async(self) -> None:
        pass

    @override
    async def shutdown_async(self) -> None:
        pass

    @contextmanager
    @override
    def trace(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> Iterator[TelemetrySpanWrapper]:
        yield NullTelemetrySpanWrapper(
            name=name,
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @asynccontextmanager
    @override
    async def trace_async(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> AsyncIterator[TelemetrySpanWrapper]:
        yield NullTelemetrySpanWrapper(
            name=name,
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def get_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        add_metadata: Optional[List[str]] = None,
    ) -> TelemetryCounter:
        """
        Get a counter metric
        :param name: Name of the counter
        :param unit: Unit of the counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryCounter(
            counter=NoOpCounter(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def get_up_down_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryUpDownCounter:
        """
        Get an up_down_counter metric

        :param name: Name of the up_down_counter
        :param unit: Unit of the up_down_counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryUpDownCounter(
            counter=NoOpUpDownCounter(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def get_histogram(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryHistogram:
        """
        Get a histograms metric

        :param name: Name of the histograms
        :param unit: Unit of the histograms
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        return TelemetryHistogram(
            histogram=NoOpHistogram(
                name=name,
                unit=unit,
                description=description,
            ),
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )



import logging
import os
import socket
import uuid
from contextlib import asynccontextmanager, contextmanager
from logging import Logger
from typing import (
    Any,
    Dict,
    Optional,
    override,
    Tuple,
    Iterator,
    AsyncIterator,
    Union,
    ClassVar,
    List,
    Mapping,
)

from opentelemetry import trace, metrics
from opentelemetry.context import Context
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.metrics import Meter, Counter, UpDownCounter, Histogram
from opentelemetry.sdk.metrics import (
    MeterProvider,
)
from opentelemetry.sdk.metrics.export import ConsoleMetricExporter
from opentelemetry.sdk.metrics.export import (
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.trace import SpanContext, NonRecordingSpan, TraceFlags, Span

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.metrics.telemetry_counter import TelemetryCounter
from helixtelemetry.telemetry.metrics.telemetry_histogram_counter import (
    TelemetryHistogram,
)
from helixtelemetry.telemetry.metrics.telemetry_up_down_counter import (
    TelemetryUpDownCounter,
)
from helixtelemetry.telemetry.providers.telemetry import Telemetry
from helixtelemetry.telemetry.spans.open_telemetry_span import OpenTelemetrySpanWrapper
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
    TelemetryAttributeValueWithoutNone,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent
from helixtelemetry.telemetry.structures.telemetry_tracers import TelemetryTracer
from helixtelemetry.telemetry.utilities.mapping_appender import (
    remove_null_values,
    append_mappings,
)


class OpenTelemetry(Telemetry):
    """
    Comprehensive OpenTelemetry instrumentation
    """

    telemetry_provider: ClassVar[str] = "OpenTelemetry"

    _trace_provider: ClassVar[Optional[TracerProvider]] = None

    _meter_provider: ClassVar[Optional[MeterProvider]] = None

    _counters: ClassVar[Dict[str, TelemetryCounter]] = {}

    _up_down_counters: ClassVar[Dict[str, TelemetryUpDownCounter]] = {}

    _histograms: ClassVar[Dict[str, TelemetryHistogram]] = {}

    def __init__(
        self,
        *,
        telemetry_context: TelemetryContext,
        log_level: Optional[Union[int, str]],
        write_telemetry_to_console: Optional[bool] = None,
    ) -> None:
        """
        Initialize OpenTelemetry tracer and instrumentation

        """
        super().__init__(
            telemetry_context=telemetry_context,
            log_level=log_level,
        )
        # Unique instance identifier
        self._instance_id = str(uuid.uuid4())

        self._logger: Logger = logging.getLogger(
            __name__,
        )
        if log_level:
            # get_logger sets the log level to the environment variable LOGLEVEL if it exists
            self._logger.setLevel(log_level)

        hostname: str = socket.gethostname()

        self._metadata: Dict[str, TelemetryAttributeValue] = {
            "service.name": self._telemetry_context.service_name,
            "deployment.environment": self._telemetry_context.environment,
            "host.name": hostname,
            "instance.id": self._instance_id,
            "instance.name": self._telemetry_context.instance_name,
        }

        if self._telemetry_context.attributes:
            self._metadata.update(
                remove_null_values(self._telemetry_context.attributes)
            )

        # Create a resource with service details
        # from https://opentelemetry.io/docs/specs/semconv/resource/
        resource_attributes = {
            ResourceAttributes.SERVICE_NAME: self._telemetry_context.service_name,
            ResourceAttributes.DEPLOYMENT_ENVIRONMENT: self._telemetry_context.environment,
            ResourceAttributes.SERVICE_INSTANCE_ID: self._telemetry_context.instance_name,
            ResourceAttributes.SERVICE_NAMESPACE: self._telemetry_context.service_namespace,
            ResourceAttributes.HOST_NAME: hostname,
        }
        image_version: Optional[str] = os.getenv("DOCKER_IMAGE_VERSION")
        if image_version:
            resource_attributes[ResourceAttributes.SERVICE_VERSION] = image_version
            resource_attributes[ResourceAttributes.CONTAINER_IMAGE_TAG] = image_version

        resource = Resource.create(resource_attributes)

        write_telemetry_to_console = write_telemetry_to_console or bool(
            os.getenv("TELEMETRY_WRITE_TO_CONSOLE", False)
        )

        # if the tracer is not setup then set it up
        if OpenTelemetry._trace_provider is None:
            otel_exporter_otlp_endpoint: Optional[str] = os.getenv(
                "otel_exporter_otlp_endpoint"
            )
            self._logger.debug(
                f"Setting up tracing on {hostname}"
                f" with otel_exporter_otlp_endpoint: {otel_exporter_otlp_endpoint}"
                f" telemetry_context.tracer_endpoint: {self._telemetry_context.tracer_endpoint}"
            )

            self.setup_tracing(
                resource=resource,
                write_telemetry_to_console=write_telemetry_to_console,
                telemetry_context=self._telemetry_context,
            )
            # see if the tracers are defined in an environment variable
            telemetry_tracers_text: Optional[str] = os.getenv("TELEMETRY_TRACERS")
            telemetry_tracers: List[TelemetryTracer] = (
                [
                    TelemetryTracer(tracer.strip())
                    for tracer in telemetry_tracers_text.split(",")
                ]
                if telemetry_tracers_text
                else []
            )
            if self._telemetry_context.trace_all_calls:
                telemetry_tracers.extend(self._telemetry_context.trace_all_calls)

        if OpenTelemetry._meter_provider is None:
            self.setup_meters(
                resource=resource,
                write_telemetry_to_console=write_telemetry_to_console,
                telemetry_context=self._telemetry_context,
            )

    # noinspection PyMethodMayBeStatic
    def setup_tracing(
        self,
        *,
        resource: Resource,
        telemetry_context: TelemetryContext,
        write_telemetry_to_console: bool,
    ) -> None:
        """
        Set up the OpenTelemetry tracer and exporter

        :param resource: Resource
        :param telemetry_context: Telemetry context
        :param write_telemetry_to_console: Whether to write telemetry to console
        """
        # Create trace provider
        OpenTelemetry._trace_provider = TracerProvider(resource=resource)
        # Create OTLP exporter
        otlp_exporter = OTLPSpanExporter(endpoint=telemetry_context.tracer_endpoint)
        # Add batch span processor
        span_processor = BatchSpanProcessor(span_exporter=otlp_exporter)
        OpenTelemetry._trace_provider.add_span_processor(span_processor)
        if write_telemetry_to_console:
            console_processor = BatchSpanProcessor(ConsoleSpanExporter())
            OpenTelemetry._trace_provider.add_span_processor(console_processor)
        # Set the global tracer provider
        trace.set_tracer_provider(OpenTelemetry._trace_provider)
        # Start instrumentation
        # self._start_instrumentation()

    # noinspection PyMethodMayBeStatic
    def setup_meters(
        self,
        *,
        resource: Resource,
        telemetry_context: TelemetryContext,
        write_telemetry_to_console: bool,
    ) -> None:
        reader = PeriodicExportingMetricReader(
            OTLPMetricExporter(endpoint=telemetry_context.metrics_endpoint),
        )

        # Add console exporter for debugging
        if write_telemetry_to_console:
            console_reader = PeriodicExportingMetricReader(
                ConsoleMetricExporter(), export_interval_millis=5000
            )
            OpenTelemetry._meter_provider = MeterProvider(
                metric_readers=[reader, console_reader], resource=resource
            )
        else:
            OpenTelemetry._meter_provider = MeterProvider(
                metric_readers=[reader], resource=resource
            )

        metrics.set_meter_provider(OpenTelemetry._meter_provider)

    def __getstate__(self) -> Dict[str, Any]:
        # Exclude certain properties from being pickled otherwise they cause errors in pickling
        return {
            k: v
            for k, v in self.__dict__.items()
            if k in ["_instance_id", "_metadata", "telemetry_context"]
        }

    # def _start_instrumentation(self) -> None:
    #     """
    #     Start system and logging instrumentation
    #     """
    #     assert OpenTelemetry._system_metrics_instrumentor is not None
    #     try:
    #         OpenTelemetry._system_metrics_instrumentor.instrument()
    #     except Exception as e:
    #         self._logger.exception(e)

    @contextmanager
    @override
    def trace(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> Iterator[TelemetrySpanWrapper]:
        """
        Create a traced context with optional parent trace linking
        """
        combined_attributes: Mapping[str, TelemetryAttributeValueWithoutNone] = (
            append_mappings(
                [
                    self._metadata,
                    telemetry_parent.attributes if telemetry_parent else {},
                    attributes,
                ]
            )
        )

        ctx: Optional[Context] = None

        if telemetry_parent and telemetry_parent.trace_id and telemetry_parent.span_id:
            # if there is not already a span then create one
            current_span: Span = trace.get_current_span()
            if current_span == trace.INVALID_SPAN:
                # Convert hex string to int, defaulting to 0 if conversion fails
                trace_id_int = (
                    int(telemetry_parent.trace_id, 16)
                    if telemetry_parent.trace_id
                    else 0
                )
                span_id_int = (
                    int(telemetry_parent.span_id, 16) if telemetry_parent.span_id else 0
                )

                # Create a SpanContext for the parent trace
                span_context = SpanContext(
                    trace_id=trace_id_int,
                    span_id=span_id_int,
                    is_remote=True,
                    trace_flags=TraceFlags(0x01),
                )
                self._logger.debug(
                    f"OpenTelemetry {self._instance_id} trace_async created span wrapper for {name}"
                    f" with trace ID: {telemetry_parent.trace_id} and span ID: {telemetry_parent.span_id}"
                )
                ctx = trace.set_span_in_context(NonRecordingSpan(span_context))
        else:
            self._logger.debug(
                f"OpenTelemetry {self._instance_id} trace_async created span wrapper for {name} without parent trace"
            )

        _tracer = trace.get_tracer(
            instrumenting_module_name=self._telemetry_context.service_name,
            tracer_provider=OpenTelemetry._trace_provider,
        )
        span: Span = _tracer.start_span(
            name=name,
            attributes=combined_attributes,
            context=ctx,
            start_time=start_time,
        )
        with trace.use_span(span=span, end_on_exit=True):
            yield OpenTelemetrySpanWrapper(
                name=name,
                attributes=combined_attributes,
                span=span,
                telemetry_parent=telemetry_parent,
            )

    @asynccontextmanager
    @override
    async def trace_async(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> AsyncIterator[TelemetrySpanWrapper]:
        """
        Async version of trace with parent trace support

        :param name: Name of the span
        :param attributes: Attributes to add to the span
        :param telemetry_parent: Parent telemetry context
        :param start_time: (Optional) Start time of the span
        :return: Async context manager
        """
        combined_attributes: Mapping[str, TelemetryAttributeValueWithoutNone] = (
            append_mappings(
                [
                    self._metadata,
                    telemetry_parent.attributes if telemetry_parent else {},
                    attributes,
                ]
            )
        )

        ctx: Optional[Context] = None

        if telemetry_parent and telemetry_parent.trace_id and telemetry_parent.span_id:
            # if there is not already a span then create one
            current_span: Span = trace.get_current_span()
            if current_span == trace.INVALID_SPAN:
                # Convert hex string to int, defaulting to 0 if conversion fails
                trace_id_int = (
                    int(telemetry_parent.trace_id, 16)
                    if telemetry_parent.trace_id
                    else 0
                )
                span_id_int = (
                    int(telemetry_parent.span_id, 16) if telemetry_parent.span_id else 0
                )

                # Create a SpanContext for the parent trace
                span_context = SpanContext(
                    trace_id=trace_id_int,
                    span_id=span_id_int,
                    is_remote=True,
                    trace_flags=TraceFlags(0x01),
                )
                self._logger.debug(
                    f"OpenTelemetry {self._instance_id} trace_async created span wrapper for {name}"
                    f" with trace ID: {telemetry_parent.trace_id} and span ID: {telemetry_parent.span_id}"
                )
                ctx = trace.set_span_in_context(NonRecordingSpan(span_context))
        else:
            self._logger.debug(
                f"OpenTelemetry {self._instance_id} trace_async created span wrapper for {name} without parent trace"
            )

        _tracer = trace.get_tracer(
            instrumenting_module_name=self._telemetry_context.service_name,
            tracer_provider=OpenTelemetry._trace_provider,
        )

        span: Span = _tracer.start_span(
            name=name,
            attributes=combined_attributes,
            context=ctx,
            start_time=start_time,
        )
        with trace.use_span(span=span, end_on_exit=True):
            yield OpenTelemetrySpanWrapper(
                name=name,
                attributes=combined_attributes,
                span=span,
                telemetry_parent=telemetry_parent,
            )

    @override
    def track_exception(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Track and record exceptions

        Args:
            exception: Exception to track
            additional_info: Extra context for the exception
        """
        current_span = trace.get_current_span()
        if current_span:
            current_span.record_exception(exception, attributes=additional_info or {})

    @override
    async def track_exception_async(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Track and record exceptions

        Args:
            exception: Exception to track
            additional_info: Extra context for the exception
        """
        current_span = trace.get_current_span()
        if current_span:
            current_span.record_exception(exception, attributes=additional_info or {})

    # noinspection PyMethodMayBeStatic
    def add_event(
        self,
        event_name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> None:
        """
        Add a custom event to the current span

        Args:
            event_name: Name of the event
            attributes: Event attributes
        """
        current_span = trace.get_current_span()
        if current_span:
            current_span.add_event(
                event_name,
                attributes=remove_null_values(attributes) if attributes else {},
            )

    async def shutdown_async(self) -> None:
        """
        Gracefully shutdown the tracer provider
        """
        if OpenTelemetry._trace_provider is not None:
            OpenTelemetry._trace_provider.force_flush()
            OpenTelemetry._trace_provider.shutdown()
            OpenTelemetry._trace_provider = None

        if OpenTelemetry._meter_provider is not None:
            OpenTelemetry._meter_provider.force_flush()
            OpenTelemetry._meter_provider.shutdown()
            OpenTelemetry._meter_provider = None

    @override
    async def flush_async(self) -> None:
        """
        Flush the span processor
        """
        if OpenTelemetry._trace_provider is not None:
            OpenTelemetry._trace_provider.force_flush()
        if OpenTelemetry._meter_provider is not None:
            OpenTelemetry._meter_provider.force_flush()

    # noinspection PyMethodMayBeStatic
    def get_current_trace_context(self) -> Optional[Tuple[str, str]]:
        """
        Retrieve the current trace and span IDs as hex strings.

        Returns:
            Optional tuple of (trace_id, span_id) in hex format,
            or None if no active trace context exists.
        """
        current_span = trace.get_current_span()

        # Check if the current span is a non-recording or invalid span
        if (
            current_span is None
            or current_span is trace.INVALID_SPAN
            or isinstance(current_span, NonRecordingSpan)
        ):
            return None

        # Get the span context
        span_context = current_span.get_span_context()

        # Validate span context
        if not span_context or not span_context.trace_id or not span_context.span_id:
            return None

        # Convert trace and span IDs to hex strings
        try:
            trace_id_hex = f"{span_context.trace_id:032x}"
            span_id_hex = f"{span_context.span_id:016x}"
            return trace_id_hex, span_id_hex
        except Exception as e:
            print(f"Error converting trace/span IDs: {e}")
            return None

    def get_current_trace_id(self) -> Optional[str]:
        """
        Retrieve the current trace ID as a hex string.

        Returns:
            Trace ID in hex format, or None if no active trace context exists.
        """
        context: Tuple[str, str] | None = self.get_current_trace_context()
        return context[0] if context else None

    def get_current_span_id(self) -> Optional[str]:
        """
        Retrieve the current span ID as a hex string.

        Returns:
            Span ID in hex format, or None if no active trace context exists.
        """
        context: Tuple[str, str] | None = self.get_current_trace_context()
        return context[1] if context else None

    @override
    def get_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        add_metadata: Optional[List[str]] = None,
    ) -> TelemetryCounter:
        """
        Get a counter metric

        :param name: Name of the counter
        :param unit: Unit of the counter
        :param description: Description
        :param attributes: Additional attributes
        :param telemetry_parent: Parent telemetry context
        :return: The Counter metric
        """
        # check if we already have a counter for this name
        if name in OpenTelemetry._counters:
            return OpenTelemetry._counters[name]

        combined_attributes: Mapping[str, TelemetryAttributeValueWithoutNone] = (
            append_mappings(
                [
                    (
                        {
                            key: self._metadata[key]
                            for key in add_metadata
                            if self._metadata.get(key)
                        }
                        if add_metadata
                        else {}
                    ),
                    telemetry_parent.attributes if telemetry_parent else {},
                    attributes,
                ]
            )
        )

        meter: Meter = metrics.get_meter(
            name=self._telemetry_context.service_name,
            meter_provider=OpenTelemetry._meter_provider,
            attributes=combined_attributes,
        )

        counter: Counter = meter.create_counter(
            name=name,
            unit=unit,
            description=description,
        )

        counter_wrapper: TelemetryCounter = TelemetryCounter(
            counter=counter,
            attributes=combined_attributes,
            telemetry_parent=telemetry_parent,
        )
        # add to the dictionary of counters
        OpenTelemetry._counters[name] = counter_wrapper

        return counter_wrapper

    @override
    def get_up_down_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryUpDownCounter:
        """
        Get an up_down_counter metric

        :param name: Name of the up_down_counter
        :param unit: Unit of the up_down_counter
        :param description: Description
        :param attributes: Additional attributes
        :param telemetry_parent: Parent telemetry context
        :return: The Counter metric
        """
        # check if we already have an up_down_counter for this name
        if name in OpenTelemetry._up_down_counters:
            return OpenTelemetry._up_down_counters[name]

        combined_attributes: Mapping[str, TelemetryAttributeValueWithoutNone] = (
            append_mappings(
                [
                    self._metadata,
                    telemetry_parent.attributes if telemetry_parent else {},
                    attributes,
                ]
            )
        )

        meter: Meter = metrics.get_meter(
            name=self._telemetry_context.service_name,
            meter_provider=OpenTelemetry._meter_provider,
            attributes=combined_attributes,
        )

        up_down_counter: UpDownCounter = meter.create_up_down_counter(
            name=name,
            unit=unit,
            description=description,
        )

        up_down_counter_wrapper: TelemetryUpDownCounter = TelemetryUpDownCounter(
            counter=up_down_counter,
            attributes=combined_attributes,
            telemetry_parent=telemetry_parent,
        )
        # add to the dictionary of counters
        OpenTelemetry._up_down_counters[name] = up_down_counter_wrapper

        return up_down_counter_wrapper

    @override
    def get_histogram(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryHistogram:
        """
        Get a histogram metric

        :param name: Name of the histogram
        :param unit: Unit of the histogram
        :param description: Description
        :param attributes: Additional attributes
        :param telemetry_parent: Parent telemetry context
        :return: The Counter metric
        """
        # check if we already have a histogram for this name
        if name in OpenTelemetry._histograms:
            return OpenTelemetry._histograms[name]

        combined_attributes: Mapping[str, TelemetryAttributeValueWithoutNone] = (
            append_mappings(
                [
                    self._metadata,
                    telemetry_parent.attributes if telemetry_parent else {},
                    attributes,
                ]
            )
        )

        meter: Meter = metrics.get_meter(
            name=self._telemetry_context.service_name,
            meter_provider=OpenTelemetry._meter_provider,
            attributes=combined_attributes,
        )

        histogram: Histogram = meter.create_histogram(
            name=name,
            unit=unit,
            description=description,
        )

        histogram_wrapper: TelemetryHistogram = TelemetryHistogram(
            histogram=histogram,
            attributes=combined_attributes,
            telemetry_parent=telemetry_parent,
        )
        # add to the dictionary of counters
        OpenTelemetry._histograms[name] = histogram_wrapper

        return histogram_wrapper



from typing import Optional, override, Any, Dict, Mapping

from opentelemetry.trace import Span

from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class OpenTelemetrySpanWrapper(TelemetrySpanWrapper):
    def __init__(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]],
        span: Span,
        telemetry_parent: Optional[TelemetryParent],
    ) -> None:
        super().__init__(
            name=name,
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )
        self._span: Span = span

    @override
    @property
    def trace_id(self) -> Optional[str]:
        span_context = self._span.get_span_context()
        if not span_context or not span_context.trace_id or not span_context.span_id:
            return None
        trace_id_hex = f"{span_context.trace_id:032x}"
        return trace_id_hex

    @override
    @property
    def span_id(self) -> Optional[str]:
        span_context = self._span.get_span_context()
        if not span_context or not span_context.trace_id or not span_context.span_id:
            return None
        span_id_hex = f"{span_context.span_id:016x}"
        return span_id_hex

    @override
    def set_attributes(self, attributes: Dict[str, Any]) -> None:
        self._span.set_attributes(attributes=attributes)

    @override
    def end(self, *, end_time: int) -> None:
        self._span.end(end_time=end_time)



from typing import Any, Dict, Optional, Type, Callable

from helixtelemetry.telemetry.providers.telemetry import Telemetry
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class TelemetryFactory:

    _registry: Dict[str, type[Telemetry]] = {}

    def __init__(self, *, telemetry_parent: TelemetryParent) -> None:
        """
        Telemetry factory used to create telemetry instances based on the telemetry context


        :param telemetry_parent: telemetry parent
        """
        self.telemetry_parent = telemetry_parent

    @classmethod
    def register_telemetry_class(
        cls, *, name: str, telemetry_class: Type[Telemetry]
    ) -> None:
        registration_name = name

        # Validate that the class is a Telemetry subclass
        if not issubclass(telemetry_class, Telemetry):
            raise TypeError(
                f"{telemetry_class.__name__} must be a subclass of Telemetry"
            )

        # Register the class in the factory's registry
        cls._registry[registration_name] = telemetry_class

    @classmethod
    def register_telemetry(
        cls, name: str
    ) -> Callable[[Type[Telemetry]], Type[Telemetry]]:
        """
        Decorator to register Telemetry subclasses in the factory registry.

        Args:
            name (Optional[str], optional): Custom registration name.
                Defaults to the class name if not provided.

        Returns:
            Callable: A decorator function for registering Telemetry classes
        """

        def decorator(telemetry_class: Type[Telemetry]) -> Type[Telemetry]:
            # Use provided name or fallback to class name
            registration_name = name

            # Validate that the class is a Telemetry subclass
            if not issubclass(telemetry_class, Telemetry):
                raise TypeError(
                    f"{telemetry_class.__name__} must be a subclass of Telemetry"
                )

            # Register the class in the factory's registry
            cls._registry[registration_name] = telemetry_class

            return telemetry_class

        return decorator

    def create(self, *, log_level: Optional[str | int]) -> Telemetry:
        """
        Create a telemetry instance

        :return: telemetry instance
        """
        assert (
            self.telemetry_parent.telemetry_context.provider in self._registry
        ), f"Telemetry {self.telemetry_parent.telemetry_context.provider} not found in registry.  Did you register a class for it using register_telemetry_class()?"
        return self._registry[self.telemetry_parent.telemetry_context.provider](
            telemetry_context=self.telemetry_parent.telemetry_context,
            log_level=log_level,
        )

    def create_telemetry_span_creator(
        self, *, log_level: Optional[str | int]
    ) -> TelemetrySpanCreator:
        """
        Create a telemetry span creator

        :return: telemetry span creator
        """
        return TelemetrySpanCreator(
            telemetry=self.create(log_level=log_level),
        )

    # noinspection PyTypeChecker
    def __getstate__(self) -> Dict[str, Any]:
        raise NotImplementedError(
            "Serialization of TelemetrySpanCreator is not supported.  Did you accidentally try to send this object to a Spark worker?"
        )



from typing import (
    Optional,
    Dict,
    Any,
    override,
    Mapping,
)

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class ConsoleTelemetrySpanWrapper(TelemetrySpanWrapper):
    @override
    @property
    def span_id(self) -> Optional[str]:
        return (
            self._telemetry_parent.span_id
            if self._telemetry_parent is not None
            else None
        )

    @override
    @property
    def trace_id(self) -> Optional[str]:
        return (
            self._telemetry_parent.trace_id
            if self._telemetry_parent is not None
            else None
        )

    def __init__(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]],
        telemetry_context: Optional[TelemetryContext],
        telemetry_parent: Optional[TelemetryParent],
    ) -> None:
        super().__init__(
            name=name,
            attributes=attributes,
            telemetry_parent=telemetry_parent,
        )

    @override
    def set_attributes(self, attributes: Dict[str, Any]) -> None:
        pass

    @override
    def end(self, *, end_time: int) -> None:
        pass



from enum import Enum


class TelemetryTracer(Enum):
    ASYNCIO = "asyncio"
    AIOHTTP = "aiohttp"
    BOTO = "boto"
    BOTOCORE = "botocore"
    CONFLUENT_KAFKA = "confluent_kafka"
    ELASTICSEARCH = "elasticsearch"
    HTTPX = "httpx"
    PYMONGO = "pymongo"
    REQUESTS = "requests"
    SYSTEM = "system"



from abc import abstractmethod, ABC
from contextlib import asynccontextmanager, contextmanager

from typing import Optional, Dict, Any, AsyncIterator, Iterator, Mapping, Union, List

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.metrics.telemetry_counter import TelemetryCounter
from helixtelemetry.telemetry.metrics.telemetry_histogram_counter import (
    TelemetryHistogram,
)
from helixtelemetry.telemetry.metrics.telemetry_up_down_counter import (
    TelemetryUpDownCounter,
)
from helixtelemetry.telemetry.spans.console_telemetry_span_wrapper import (
    ConsoleTelemetrySpanWrapper,
)
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class Telemetry(ABC):
    """
    Abstract class for telemetry

    """

    def __init__(
        self,
        *,
        telemetry_context: TelemetryContext,
        log_level: Optional[Union[int, str]],
    ) -> None:
        self._telemetry_context: TelemetryContext = telemetry_context
        self._log_level = log_level

    @abstractmethod
    @contextmanager
    def trace(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> Iterator[TelemetrySpanWrapper]:
        """
        Start a new span

        :param name:  name of the span
        :param attributes: optional attributes to add to the span
        :param telemetry_parent: parent span
        :param start_time: start time
        :return: A context manager to use in a `with` statement
        """
        # This is never called but is here for mypy to understand this is a generator
        yield ConsoleTelemetrySpanWrapper(
            name=name,
            attributes=attributes,
            telemetry_context=None,
            telemetry_parent=None,
        )

    @abstractmethod
    @asynccontextmanager
    async def trace_async(
        self,
        *,
        name: str,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent],
        start_time: int | None = None,
    ) -> AsyncIterator[TelemetrySpanWrapper]:
        """
        Start a new span

        :param name:  name of the span
        :param attributes: optional attributes to add to the span
        :param telemetry_parent: telemetry parent
        :param start_time: start time
        :return: A context manager to use in a `with` statement
        """
        # This is never called but is here for mypy to understand this is a generator
        yield ConsoleTelemetrySpanWrapper(
            name=name,
            attributes=attributes,
            telemetry_context=None,
            telemetry_parent=telemetry_parent,
        )

    @abstractmethod
    def track_exception(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Track and record exceptions

        :param exception: exception to track
        :param additional_info: Optional extra context for the exception
        :return: None
        """
        ...

    @abstractmethod
    async def track_exception_async(
        self, exception: Exception, additional_info: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Track and record exceptions

        :param exception: exception to track
        :param additional_info: Optional extra context for the exception
        :return: None
        """
        ...

    @abstractmethod
    async def flush_async(self) -> None: ...

    @abstractmethod
    async def shutdown_async(self) -> None: ...

    @abstractmethod
    def get_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        add_metadata: Optional[List[str]] = None,
    ) -> TelemetryCounter:
        """
        Get a counter metric

        :param name: Name of the counter
        :param unit: Unit of the counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        ...

    @abstractmethod
    def get_up_down_counter(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryUpDownCounter:
        """
        Get an up_down_counter metric

        :param name: Name of the up_down_counter
        :param unit: Unit of the up_down_counter
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        ...

    @abstractmethod
    def get_histogram(
        self,
        *,
        name: str,
        unit: str,
        description: str,
        telemetry_parent: Optional[TelemetryParent],
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
    ) -> TelemetryHistogram:
        """
        Get a histograms metric

        :param name: Name of the histograms
        :param unit: Unit of the histograms
        :param description: Description
        :param attributes: Optional attributes
        :param telemetry_parent: telemetry parent
        :return: The Counter metric
        """
        ...
